@@ -26,16 +26,22 @@ var FireGestures = {
 		}
 		var gestureSvc = Cc["@xuldev.org/firegestures/service;1"].getService(Ci.xdIGestureService);
 		this._gestureHandler = gestureSvc.createHandler();
-		this._gestureHandler.attach(gBrowser.mPanelContainer, this);
+		this._gestureHandler.attach(gBrowser, this);
 		this._gestureMapping = gestureSvc.getMappingForBrowser();
 		this._getLocaleString = gestureSvc.getLocaleString;
-		this._statusTextField = document.getElementById("statusbar-display");
-		if (!this._statusTextField)
-			this._statusTextField = gBrowser.getStatusPanel();
+		this._statusTextField = gBrowser.getStatusPanel();
 		window.removeEventListener("MozSwipeGesture", gGestureSupport, true);
+		if (gMultiProcessBrowser) {
+			window.messageManager.loadFrameScript("chrome://firegestures/content/remote.js", true);
+			window.messageManager.addMessageListener("FireGesturesRemote:Response", this);
+		}
 	},
 
 	uninit: function() {
+		if (gMultiProcessBrowser) {
+			window.messageManager.removeDelayedFrameScript("chrome://firegestures/content/remote.js");
+			window.messageManager.removeMessageListener("FireGesturesRemote:Response", this);
+		}
 		if (this._gestureHandler) {
 			this._gestureHandler.detach();
 			this._gestureHandler = null;
@@ -49,10 +55,36 @@ var FireGestures = {
 
 
 
-	get fullZoom() {
-		return gBrowser.mCurrentBrowser.markupDocumentViewer.fullZoom;
+	get isRemote() {
+		return gBrowser.mCurrentBrowser.getAttribute("remote") == "true";
 	},
 
+	sendAsyncMessage: function(aName, aData) {
+		gBrowser.mCurrentBrowser.messageManager.sendAsyncMessage(aName, aData);
+	},
+
+	receiveMessage: function(aMsg) {
+		switch (aMsg.data.name) {
+			case "sourceNode": 
+				this._gestureHandler.sourceNode = aMsg.objects.elt;
+				this._selectedText = aMsg.objects.sel;
+				break;
+			case "linkURLs": 
+				this._linkURLs = aMsg.data.linkURLs;
+				break;
+			case "cancelMouseGesture": 
+				this._gestureHandler.cancelMouseGesture();
+				break;
+			case "swipe": 
+				this.onMouseGesture(null, "swipe-" + aMsg.objects.direction);
+				this._gestureHandler.sourceNode = null;
+				break;
+			default: 
+		}
+	},
+
+
+
 	canStartGesture: function(event) {
 		if (gInPrintPreviewMode) {
 			return false;
@@ -110,11 +142,23 @@ var FireGestures = {
 				this.onMouseGesture(event, aGesture);
 				break;
 			case "keypress-start": 
+				if (this.isRemote) {
+					this._linkURLs = [];
+					this.sendAsyncMessage("FireGestures:KeypressStart", {});
+					return;
+				}
 				this.clearStatusText(0);
 				this._linkURLs = [];
 				this._linkElts = [];
 				break;
 			case "keypress-progress": 
+				if (this.isRemote) {
+					this.sendAsyncMessage("FireGestures:KeypressProgress", {
+						x: event.screenX - gBrowser.mCurrentBrowser.boxObject.screenX, 
+						y: event.screenY - gBrowser.mCurrentBrowser.boxObject.screenY, 
+					});
+					return;
+				}
 				var linkURL = this.getLinkURL(event.target);
 				if (!this._linkURLs)
 					this._linkURLs = [];
@@ -129,6 +173,11 @@ var FireGestures = {
 				}
 				break;
 			case "keypress-stop": 
+				if (this.isRemote) {
+					this._linkURLs = null;
+					this.sendAsyncMessage("FireGestures:KeypressStop", {});
+					return;
+				}
 				for (var i = 0; i < this._linkURLs.length; i++) {
 					this._linkElts[i].style.outline = "";
 					this._linkElts[i] = null;
@@ -217,19 +266,23 @@ var FireGestures = {
 			case "FireGestures:CloseRightTabs": 
 				this.closeMultipleTabs("right");
 				break;
+			case "Browser:ToggleTabView": 
+				alert("Please add the following script:\nTabView.toggle();");
+				this._gestureMapping.configure();
+				break;
 			case "cmd_textZoomEnlarge": 
 			case "cmd_textZoomReduce": 
 				if ("FullZoom" in window && !ZoomManager.useFullZoom)
 					document.getElementById(aCommand.replace("text", "full")).doCommand();
 				else
-					gBrowser.markupDocumentViewer.textZoom += (aCommand == "cmd_textZoomEnlarge") ? 0.2 : -0.2;
+					gBrowser.textZoom += (aCommand == "cmd_textZoomEnlarge") ? 0.2 : -0.2;
 				break;
 			case "cmd_fullZoomEnlarge": 
 			case "cmd_fullZoomReduce": 
 				if (ZoomManager.useFullZoom)
 					document.getElementById(aCommand).doCommand();
 				else
-					gBrowser.markupDocumentViewer.fullZoom += (aCommand == "cmd_fullZoomEnlarge") ? 0.2 : -0.2;
+					gBrowser.fullZoom += (aCommand == "cmd_fullZoomEnlarge") ? 0.2 : -0.2;
 				break;
 			case "cmd_textZoomReset": 
 				if ("FullZoom" in window)
@@ -237,43 +290,84 @@ var FireGestures = {
 				document.getElementById(aCommand).doCommand();
 				break;
 			case "FireGestures:ScrollTop": 
-				this._performScrollAction("cmd_scrollTop", "DOM_VK_HOME");
-				break;
 			case "FireGestures:ScrollBottom": 
-				this._performScrollAction("cmd_scrollBottom", "DOM_VK_END");
-				break;
 			case "FireGestures:ScrollPageUp": 
-				this._performScrollAction("cmd_scrollPageUp", "DOM_VK_PAGE_UP");
-				break;
 			case "FireGestures:ScrollPageDown": 
-				this._performScrollAction("cmd_scrollPageDown", "DOM_VK_PAGE_DOWN");
+				aCommand = aCommand.replace("FireGestures:Scroll", "cmd_scroll");
+				this.sourceNode.ownerDocument.defaultView.focus();
+				if (this.isRemote) {
+					this.sendAsyncMessage("FireGestures:DoCommand", { cmd: aCommand });
+					return;
+				}
+				var docShell = gBrowser.mCurrentBrowser.docShell;
+				if (docShell.isCommandEnabled(aCommand))
+					docShell.doCommand(aCommand);
 				break;
 			case "FireGestures:ShowOnlyThisFrame": 
-				var docURL = this.sourceNode.ownerDocument.location.href;
-				this.checkURL(docURL, gBrowser.contentDocument, Ci.nsIScriptSecurityManager.DISALLOW_SCRIPT);
-				gBrowser.loadURI(docURL);
+				var doc = this.sourceNode.ownerDocument;
+				var docURL = doc.location.href;
+				var refURI = doc.referrer ? makeURI(doc.referrer) : null;
+				this.checkURL(docURL, doc.defaultView.top.document, Ci.nsIScriptSecurityManager.DISALLOW_SCRIPT);
+				openUILinkIn(docURL, "current", { disallowInheritPrincipal: true, referrerURI: refURI });
 				break;
 			case "FireGestures:OpenFrame": 
 			case "FireGestures:OpenFrameInTab": 
+				var doc = this.sourceNode.ownerDocument;
+				var docURL = doc.location.href;
+				var refURI = doc.referrer ? makeURI(doc.referrer) : null;
+				openLinkIn(docURL, aCommand == "FireGestures:OpenFrame" ? "window" : "tab", {
+					charset: doc.characterSet, referrerURI: refURI
+				});
+				break;
 			case "FireGestures:ReloadFrame": 
+				this.sourceNode.ownerDocument.location.reload();
+				break;
 			case "FireGestures:AddBookmarkForFrame": 
+				var doc = this.sourceNode.ownerDocument;
+				var docURI = makeURI(doc.location.href);
+				var itemId = PlacesUtils.getMostRecentBookmarkForURI(docURI);
+				if (itemId == -1) {
+					PlacesUIUtils.showBookmarkDialog({
+						action: "add", type: "bookmark", uri: docURI, title: doc.title, 
+						description: PlacesUIUtils.getDescriptionFromDocument(doc), 
+						hiddenRows: ["description", "location", "loadInSidebar", "keyword" ]
+					}, window.top);
+				}
+				else {
+					PlacesUIUtils.showBookmarkDialog({
+						action: "edit", type: "bookmark", itemId: itemId
+					}, window.top);
+				}
+				break;
 			case "FireGestures:SaveFrame": 
+				saveDocument(this.sourceNode.ownerDocument);
+				break;
 			case "FireGestures:ViewFrameSource": 
+				if (this.isRemote) {
+					var doc = this.sourceNode.ownerDocument;
+					var frameID = doc.defaultView.QueryInterface(Ci.nsIInterfaceRequestor).
+					              getInterface(Ci.nsIDOMWindowUtils).outerWindowID;
+					BrowserViewSourceOfDocument({
+						browser: gBrowser.mCurrentBrowser, URL: doc.location.href, 
+						outerWindowID: frameID
+					});
+					return;
+				}
+				BrowserViewSourceOfDocument(this.sourceNode.ownerDocument);
+				break;
 			case "FireGestures:ViewFrameInfo": 
-				var funcName = aCommand.substr("FireGestures:".length);
-				funcName = funcName.charAt(0).toLowerCase() + funcName.substr(1);
-				nsContextMenu.prototype.target = this.sourceNode;
-				try { nsContextMenu.prototype[funcName](); }
-				finally { nsContextMenu.prototype.target = null; }
+				BrowserPageInfo(this.sourceNode.ownerDocument);
 				break;
 			case "FireGestures:OpenLink": 
+			case "FireGestures:OpenLinkInPrivateWindow": 
 				var linkURL = this.getLinkURL();
 				if (!linkURL)
 					throw this._getLocaleString("ERROR_NOT_ON_LINK");
 				var doc = this.sourceNode.ownerDocument;
 				this.checkURL(linkURL, doc);
 				openLinkIn(linkURL, "window", {
-					charset: doc.characterSet, referrerURI: doc.documentURIObject
+					charset: doc.characterSet, referrerURI: makeURI(doc.location.href), 
+					private: aCommand == "FireGestures:OpenLinkInPrivateWindow"
 				});
 				break;
 			case "FireGestures:OpenLinkInBgTab": 
@@ -284,21 +378,11 @@ var FireGestures = {
 				var doc = this.sourceNode.ownerDocument;
 				this.checkURL(linkURL, doc);
 				gBrowser.loadOneTab(linkURL, {
-					referrerURI: doc.documentURIObject, charset: doc.characterSet, 
+					referrerURI: makeURI(doc.location.href), charset: doc.characterSet, 
 					inBackground: aCommand == "FireGestures:OpenLinkInBgTab", 
 					relatedToCurrent: true
 				});
 				break;
-			case "FireGestures:OpenLinkInPrivateWindow": 
-				var linkURL = this.getLinkURL();
-				if (!linkURL)
-					throw this._getLocaleString("ERROR_NOT_ON_LINK");
-				var doc = this.sourceNode.ownerDocument;
-				this.checkURL(linkURL, doc);
-				openLinkIn(linkURL, "window", {
-					charset: doc.characterSet, referrerURI: doc.documentURIObject, private: true
-				});
-				break;
 			case "FireGestures:AddBookmarkForLink": 
 				var linkURL = this.getLinkURL();
 				if (!linkURL)
@@ -311,8 +395,7 @@ var FireGestures = {
 					throw this._getLocaleString("ERROR_NOT_ON_LINK");
 				var doc = this.sourceNode.ownerDocument;
 				this.checkURL(linkURL, doc);
-				var linkText = gatherTextUnder(this.sourceNode);
-				nsContextMenu.prototype.saveHelper(linkURL, linkText, null, true, doc);
+				nsContextMenu.prototype.saveHelper(linkURL, this.getLinkText(), null, true, doc);
 				break;
 			case "FireGestures:ViewImage": 
 				var imageURL = this.getImageURL();
@@ -320,7 +403,8 @@ var FireGestures = {
 					throw this._getLocaleString("ERROR_NOT_ON_IMAGE");
 				var onCanvas = this.sourceNode instanceof HTMLCanvasElement;
 				if (onCanvas)
-					this.checkURL(imageURL, gBrowser.contentDocument, Ci.nsIScriptSecurityManager.DISALLOW_SCRIPT);
+					this.checkURL(imageURL, this.sourceNode.ownerDocument, 
+					              Ci.nsIScriptSecurityManager.DISALLOW_SCRIPT);
 				openUILink(imageURL, event);
 				break;
 			case "FireGestures:SaveImage": 
@@ -330,6 +414,8 @@ var FireGestures = {
 					throw this._getLocaleString("ERROR_NOT_ON_IMAGE");
 				var doc = this.sourceNode.ownerDocument;
 				var skipPrompt = aCommand == "FireGestures:SaveImageNow";
+				var refURI = makeURI(doc.location.href);
+				var isPrivate = PrivateBrowsingUtils.isBrowserPrivate(gBrowser);
 				if (this.sourceNode instanceof HTMLVideoElement || 
 				    this.sourceNode instanceof HTMLAudioElement) {
 					this.checkURL(mediaURL, doc);
@@ -338,26 +424,28 @@ var FireGestures = {
 					nsContextMenu.prototype.saveHelper(mediaURL, null, dialogTitle, false, doc);
 				}
 				else if (this.sourceNode instanceof HTMLCanvasElement) {
-					saveImageURL(mediaURL, "canvas.png", "SaveImageTitle", 
-					             false, skipPrompt, doc.documentURIObject, doc);
+					this.isRemote ? 
+					saveImageURL(mediaURL, "canvas.png", "SaveImageTitle", false, skipPrompt, refURI, 
+					             null, null, null, isPrivate) : 
+					saveImageURL(mediaURL, "canvas.png", "SaveImageTitle", false, skipPrompt, refURI, doc);
 				}
 				else {
 					this.checkURL(mediaURL, doc);
-					saveImageURL(mediaURL, null, "SaveImageTitle", 
-					             false, skipPrompt, doc.documentURIObject, doc);
+					this.isRemote ? 
+					saveImageURL(mediaURL, null, "SaveImageTitle", false, skipPrompt, refURI, 
+					             null, null, null, isPrivate) : 
+					saveImageURL(mediaURL, null, "SaveImageTitle", false, skipPrompt, refURI, doc);
 				}
 				break;
 			case "FireGestures:WebSearch": 
-				BrowserSearch.loadSearch(getBrowserSelection(), true);
+				BrowserSearch.loadSearch(this.getSelectedText(), true);
 				break;
 			case "FireGestures:OpenLinksInSelection": 
 				var linkURLs = this.gatherLinkURLsInSelection();
 				if (!linkURLs || linkURLs.length == 0)
 					throw "No valid links in selection";
 				var doc = this.sourceNode.ownerDocument;
-				var referer = makeURI(doc.location.href);
-				var charset = window.content.document.characterSet;
-				this.openURLs(linkURLs, referer, charset);
+				this.openURLs(linkURLs, makeURI(doc.location.href), doc.charset);
 				break;
 			case "FireGestures:OpenURLsInSelection": 
 				this.openURLsInSelection();
@@ -378,16 +466,24 @@ var FireGestures = {
 				gFindBar.hidden ? gFindBar.onFindCommand() : gFindBar.close();
 				break;
 			case "FireGestures:RestartApp": 
-				Application.restart();
+				let cancelQuit = Cc["@mozilla.org/supports-PRBool;1"].
+				                 createInstance(Ci.nsISupportsPRBool);
+				Services.obs.notifyObservers(cancelQuit, "quit-application-requested", "restart");
+				if (cancelQuit.data)
+					return;
+				let appStartup = Cc["@mozilla.org/toolkit/app-startup;1"].
+				                 getService(Ci.nsIAppStartup);
+				appStartup.quit(Ci.nsIAppStartup.eAttemptQuit |  Ci.nsIAppStartup.eRestart);
 				break;
 			case "FireGestures:Preferences": 
 				this._gestureMapping.configure();
 				break;
 			case "FireGestures:HybridSave": 
 			case "FireGestures:HybridBookmark": 
+				var doc = this.sourceNode.ownerDocument;
 				var onLink  = this.getLinkURL()  != null;
 				var onMedia = this.getMediaURL() != null;
-				var inFrame = this.sourceNode.ownerDocument != window.content.document;
+				var inFrame = doc.defaultView != doc.defaultView.top;
 				if (aCommand == "FireGestures:HybridSave") {
 					if (onLink)       aCommand = "FireGestures:SaveLink";
 					else if (onMedia) aCommand = "FireGestures:SaveImage";
@@ -401,13 +497,6 @@ var FireGestures = {
 				}
 				this._performAction(event, aCommand);
 				break;
-			case "FireGestures:HybridSendURL": 
-				var url = this.getLinkURL() || this.getImageURL();
-				if (url)
-					MailIntegration.sendMessage(url, "");
-				else
-					MailIntegration.sendLinkForWindow(window.content);
-				break;
 			case "FireGestures:HybridCopyURL": 
 				var url = this.getLinkURL() || this.getImageURL() || 
 				          this.sourceNode.ownerDocument.location.href;
@@ -424,7 +513,7 @@ var FireGestures = {
 				if (this.getSelectedText())
 					nsContextMenu.prototype.viewPartialSource("selection");
 				else
-					BrowserViewSourceOfDocument(this.sourceNode.ownerDocument);
+					this._performAction(event, "FireGestures:ViewFrameSource");
 				break;
 			case "FireGestures:AllTabsPopup": 
 			case "FireGestures:BFHistoryPopup": 
@@ -458,16 +547,14 @@ var FireGestures = {
 				break;
 			case "FireGestures:OpenHoveredLinks": 
 				var doc = this.sourceNode.ownerDocument;
-				var referer = makeURI(doc.location.href);
-				var charset = window.content.document.characterSet;
-				this.openURLs(this._linkURLs, referer, charset);
+				this.openURLs(this._linkURLs, makeURI(doc.location.href), doc.characterSet);
 				break;
 			case "FireGestures:SaveHoveredLinks": 
 				var delay = 0;
 				var doc = this.sourceNode.ownerDocument;
-				var ref = makeURI(doc.location.href, doc.characterSet);
+				var refURI = makeURI(doc.location.href);
 				this._linkURLs.forEach(function(aURL) {
-					window.setTimeout(function() { saveURL(aURL, null, null, false, true, ref, doc); }, delay);
+					window.setTimeout(function() { saveURL(aURL, null, null, false, true, refURI, doc); }, delay);
 					delay += 1000;
 				});
 				break;
@@ -488,16 +575,6 @@ var FireGestures = {
 		}
 	},
 
-	_performScrollAction: function(aCommand, aKeyCode) {
-		if (!this._isWin || 
-		    this.sourceNode instanceof HTMLInputElement || 
-		    this.sourceNode instanceof HTMLTextAreaElement || 
-		    gBrowser.mPrefs.getBoolPref("accessibility.browsewithcaret"))
-			goDoCommand(aCommand);
-		else
-			this.sendKeyEvent({ keyCode: aKeyCode });
-	},
-
 	get sourceNode() {
 		return this._gestureHandler.sourceNode;
 	},
@@ -505,7 +582,7 @@ var FireGestures = {
 	get focusedWindow() {
 		var win = document.commandDispatcher.focusedWindow;
 		if (win == window)
-			win = window.content;
+			win = this.sourceNode.ownerDocument.defaultView;
 		return win;
 	},
 
@@ -523,7 +600,30 @@ var FireGestures = {
 	getLinkText: function(aNode) {
 		if (!aNode)
 			aNode = this.sourceNode;
-		var text = gatherTextUnder(aNode);
+		var text = "", node = aNode.firstChild, depth = 1;
+		while (node && depth > 0) {
+			if (node.nodeType == Node.TEXT_NODE) {
+				text += " " + node.data;
+			}
+			else if (node instanceof HTMLImageElement) {
+				var altText = node.getAttribute("alt");
+				if (altText && altText != "")
+					text += " " + altText;
+			}
+			if (node.hasChildNodes()) {
+				node = node.firstChild;
+				depth++;
+			}
+			else {
+				while (depth > 0 && !node.nextSibling) {
+					node = node.parentNode;
+					depth--;
+				}
+				if (node.nextSibling)
+					node = node.nextSibling;
+			}
+		}
+		text = text.trim().replace(/\s+/g, " ");
 		if (!text || !text.match(/\S/)) {
 			text = aNode.getAttribute("title");
 			if (!text || !text.match(/\S/)) {
@@ -571,8 +671,22 @@ var FireGestures = {
 			return this.getImageURL(aNode);
 	},
 
+	_selectedText: null,
+
 	getSelectedText: function() {
-		return this.focusedWindow.getSelection().toString();
+		if (this.isRemote) {
+			return this._selectedText;
+		}
+		var win = document.commandDispatcher.focusedWindow;
+		var elt = document.commandDispatcher.focusedElement;
+		var sel = win.getSelection().toString();
+		if (!sel && elt instanceof Ci.nsIDOMNSEditableElement) {
+			if (elt instanceof Ci.nsIDOMHTMLTextAreaElement || 
+			    (elt instanceof Ci.nsIDOMHTMLInputElement && elt.mozIsTextField(true))) {
+				sel = elt.editor.selection.toString();
+			}
+		}
+		return sel;
 	},
 
 	gatherLinkURLsInSelection: function() {
@@ -602,6 +716,15 @@ var FireGestures = {
 	},
 
 	checkURL: function(aURL, aDoc, aFlags) {
+		if (this.isRemote) {
+			let principal = Cc["@mozilla.org/scriptsecuritymanager;1"].
+			                getService(Ci.nsIScriptSecurityManager).
+			                getAppCodebasePrincipal(gBrowser.currentURI,
+			                                        aDoc.nodePrincipal.appId,
+			                                        aDoc.nodePrincipal.isInBrowserElement);
+			urlSecurityCheck(aURL, principal, aFlags);
+			return;
+		}
 		urlSecurityCheck(aURL, aDoc.nodePrincipal, aFlags);
 	},
 
@@ -683,14 +806,8 @@ var FireGestures = {
 		if (shouldPrompt && tabs.length > 1) {
 			var ps = Services.prompt;
 			var bundle = gBrowser.mStringBundle;
-			var message;
-			try {
-				message = bundle.getFormattedString("tabs.closeWarningMultipleTabs", [tabs.length]);
-			}
-			catch (ex) {
-				message = PluralForm.get(tabs.length, bundle.getString("tabs.closeWarningMultiple")).
-				          replace("#1", tabs.length);
-			}
+			var message = PluralForm.get(tabs.length, bundle.getString("tabs.closeWarningMultiple")).
+			              replace("#1", tabs.length);
 			window.focus();
 			var ret = ps.confirmEx(
 				window, bundle.getString("tabs.closeWarningTitle"), message, 
@@ -705,6 +822,10 @@ var FireGestures = {
 	},
 
 	sendKeyEvent: function(aOptions) {
+		if (this.isRemote) {
+			this.sendAsyncMessage("FireGestures:SendKeyEvent", aOptions);
+			return;
+		}
 		var evt = this.sourceNode.ownerDocument.createEvent("KeyEvents");
 		evt.initKeyEvent(
 			"keypress", true, true, null, 
@@ -806,37 +927,49 @@ var FireGestures = {
 				}
 				break;
 			case "FireGestures:BFHistoryPopup": 
-				var sessionHistory = gBrowser.webNavigation.sessionHistory;
-				if (sessionHistory.count < 1)
-					throw "No back/forward history for this tab.";
-				var curIdx = sessionHistory.index;
-				for (var i = 0; i < sessionHistory.count; i++) {
-					let entry = sessionHistory.getEntryAtIndex(i, false);
-					if (!entry)
-						continue;
-					let menuitem = document.createElement("menuitem");
-					popup.insertBefore(menuitem, popup.firstChild);
-					menuitem.setAttribute("label", entry.title);
-					menuitem.setAttribute("statustext", entry.URI.spec);
-					menuitem.index = i;
-					if (i == curIdx) {
-						menuitem.setAttribute("type", "radio");
-						menuitem.setAttribute("checked", "true");
-						menuitem.setAttribute("default", "true");
-						menuitem.className = "unified-nav-current";
-					}
-					else {
-						PlacesUtils.favicons.getFaviconURLForPage(entry.URI, function(aURI) {
-							if (!aURI)
-								return;
-							let iconURL = PlacesUtils.favicons.getFaviconLinkForIcon(aURI).spec;
-							menuitem.style.listStyleImage = "url(" + iconURL + ")";
-						});
-						menuitem.className = i < curIdx
-						                   ? "unified-nav-back menuitem-iconic menuitem-with-favicon"
-						                   : "unified-nav-forward menuitem-iconic menuitem-with-favicon";
+				function callback(sessionHistory, initial) {
+					if (popup.firstChild)
+						return;
+					let count = sessionHistory.entries ? sessionHistory.entries.length
+					                                   : sessionHistory.count;
+					if (count < 1)
+						throw "No back/forward history for this tab.";
+					var curIdx = sessionHistory.index;
+					for (var i = 0; i < count; i++) {
+						let entry = sessionHistory.entries ? sessionHistory.entries[i]
+						                                   : sessionHistory.getEntryAtIndex(i, false);
+						let menuitem = document.createElement("menuitem");
+						popup.insertBefore(menuitem, popup.firstChild);
+						menuitem.setAttribute("label", entry.title);
+						menuitem.setAttribute("statustext", entry.url);
+						menuitem.index = i;
+						if (i == curIdx) {
+							menuitem.setAttribute("type", "radio");
+							menuitem.setAttribute("checked", "true");
+							menuitem.setAttribute("default", "true");
+							menuitem.className = "unified-nav-current";
+						}
+						else {
+							var entryURI = entry.url ? BrowserUtils.makeURI(entry.url, entry.charset, null)
+							                         : this.isRemote ? makeURI(entry.URI.spec) : entry.URI;
+							PlacesUtils.favicons.getFaviconURLForPage(entryURI, function(aURI) {
+								if (!aURI)
+									return;
+								let iconURL = PlacesUtils.favicons.getFaviconLinkForIcon(aURI).spec;
+								menuitem.style.listStyleImage = "url(" + iconURL + ")";
+							});
+							menuitem.className = i < curIdx
+							                   ? "unified-nav-back menuitem-iconic menuitem-with-favicon"
+							                   : "unified-nav-forward menuitem-iconic menuitem-with-favicon";
+						}
 					}
+					if (popup.state == "showing" || popup.state == "open")
+						setTimeout(function() { popup.currentItem = popup.defaultItem || popup.firstChild; }, 10);
 				}
+				if (SessionStore.getSessionHistory)
+					SessionStore.getSessionHistory(gBrowser.selectedTab, callback);
+				else
+					callback(gBrowser.webNavigation.sessionHistory);
 				break;
 			case "FireGestures:ClosedTabsPopup": 
 				var ss = Cc["@mozilla.org/browser/sessionstore;1"].getService(Ci.nsISessionStore);
@@ -935,7 +1068,7 @@ var FireGestures = {
 						var engine = item.engine;
 						if (!engine)
 							break;
-						var submission = engine.getSubmission(getBrowserSelection(), null);
+						var submission = engine.getSubmission(this.getSelectedText(), null);
 						if (!submission)
 							break;
 						gBrowser.loadOneTab(submission.uri.spec, {
