@@ -48,6 +48,8 @@ xdGestureHandler.prototype = {
 
 	_gestureObserver: null,
 
+	_isRemote: false,
+
 	attach: function FGH_attach(aDrawArea, aObserver) {
 		var appInfo = Cc["@mozilla.org/xre/app-info;1"].getService(Ci.nsIXULAppInfo);
 		this._drawArea = aDrawArea;
@@ -79,6 +81,8 @@ xdGestureHandler.prototype = {
 		}
 		this.sourceNode = null;
 		this._drawArea = null;
+		this._trailArea = null;
+		this._trailContext = null;
 		this._gestureObserver = null;
 	},
 
@@ -120,11 +124,10 @@ xdGestureHandler.prototype = {
 			this._drawArea.addEventListener("DOMMouseScroll", this, true);
 		if (this._rockerGestureEnabled)
 			this._drawArea.addEventListener("click", this, true);
-		var tabbrowser = this._drawArea.ownerDocument.getBindingParent(this._drawArea);
-		if (tabbrowser && tabbrowser.localName == "tabbrowser") {
-			tabbrowser.mStrip.removeEventListener("DOMMouseScroll", this._wheelOnTabBar, true);
+		if (this._drawArea.localName == "tabbrowser") {
+			this._drawArea.tabContainer.removeEventListener("wheel", this._wheelOnTabBar, true);
 			if (getPref("tabwheelgesture"))
-				tabbrowser.mStrip.addEventListener("DOMMouseScroll", this._wheelOnTabBar, true);
+				this._drawArea.tabContainer.addEventListener("wheel", this._wheelOnTabBar, true);
 		}
 		if (this._triggerButton == 1) {
 			var prefSvc = Cc["@mozilla.org/preferences-service;1"]
@@ -152,23 +155,22 @@ xdGestureHandler.prototype = {
 				if (!this._gestureObserver.canStartGesture(event))
 					break;
 				if (event.button == 0) {
-					var targetName = event.target.localName.toUpperCase();
-					if (targetName == "INPUT" || targetName == "TEXTAREA") {
-						break;
-					}
-					targetName = event.originalTarget.localName;
-					if (targetName == "scrollbarbutton" || targetName == "slider" || targetName == "thumb") {
-						break;
-					}
-					if (this._triggerButton == 0 && targetName == "select") {
-						break;
+					if (this._triggerButton == 0) {
+						var localName = event.target.localName;
+						if (["input", "textarea", "select", "option", "textbox", "menulist"].indexOf(localName) >= 0) {
+							break;
+						}
+						var localName = event.originalTarget.localName;
+						if (["scrollbarbutton", "slider", "thumb"].indexOf(localName) >= 0) {
+							break;
+						}
 					}
 					this._isMouseDownL = true;
 					this._isMouseDownM = false;
 					if (this._triggerButton == 0 && !this._isMouseDownM && !this._isMouseDownR && !this._altKey(event)) {
 						this._state = STATE_GESTURE;
 						this._startGesture(event);
-						if (this._mouseGestureEnabled)
+						if (!this._isRemote && this._mouseGestureEnabled)
 							event.preventDefault();
 					}
 					else if (this._rockerGestureEnabled && this._isMouseDownR) {
@@ -181,12 +183,11 @@ xdGestureHandler.prototype = {
 					if (this._triggerButton == 1 && !this._isMouseDownL && !this._isMouseDownR && !this._altKey(event)) {
 						this._state = STATE_GESTURE;
 						this._startGesture(event);
-						event.stopPropagation();
 					}
 				}
 				else if (event.button == 2) {
-					var targetName = event.target.localName.toUpperCase();
-					if (targetName == "OBJECT" || targetName == "EMBED") {
+					var localName = event.target.localName;
+					if (localName == "object" || localName == "embed") {
 						break;
 					}
 					this._isMouseDownR = true;
@@ -212,6 +213,10 @@ xdGestureHandler.prototype = {
 							this._invokeExtraGesture(event, type);
 						}
 						this._progressGesture(event);
+						if (this._triggerButton == 1 && this._isMouseDownM && 
+						    this._drawArea.mCurrentBrowser._autoScrollPopup) {
+							this._drawArea.mCurrentBrowser._autoScrollPopup.hidePopup();
+						}
 					}
 				}
 				else if (this._state == STATE_WHEEL || this._state == STATE_ROCKER) {
@@ -242,7 +247,12 @@ xdGestureHandler.prototype = {
 					this._stopGesture(event);
 					if (this._shouldFireContext) {
 						this._shouldFireContext = false;
-						this._displayContextMenu(event);
+						this._enableContextMenu(true);
+						var win = this._drawArea.ownerDocument.defaultView;
+						var x = event.screenX - win.document.documentElement.boxObject.screenX;
+						var y = event.screenY - win.document.documentElement.boxObject.screenY;
+						win.QueryInterface(Ci.nsIInterfaceRequestor).getInterface(Ci.nsIDOMWindowUtils).
+						    sendMouseEvent("contextmenu", x, y, 2, 1, null);
 					}
 				}
 				break;
@@ -291,9 +301,17 @@ xdGestureHandler.prototype = {
 						case event.DIRECTION_UP   : direction = "up";    break;
 						case event.DIRECTION_DOWN : direction = "down";  break;
 					}
+					this._isRemote = this._drawArea.mCurrentBrowser.getAttribute("remote") == "true";
+					if (this._isRemote) {
+						var zoom = this._drawArea.mCurrentBrowser.fullZoom;
+						this._gestureObserver.sendAsyncMessage("FireGestures:SwipeGesture", {
+							direction: direction, 
+							x: (event.screenX - this._drawArea.mCurrentBrowser.boxObject.screenX) / zoom, 
+							y: (event.screenY - this._drawArea.mCurrentBrowser.boxObject.screenY) / zoom, 
+						});
+						return;
+					}
 					this.sourceNode = event.target;
-					this._lastX = event.screenX;
-					this._lastY = event.screenY;
 					this._invokeExtraGesture(event, "swipe-" + direction);
 					this.sourceNode = null;
 					return;
@@ -305,9 +323,7 @@ xdGestureHandler.prototype = {
 				this._swipeTimer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
 				this._swipeTimer.initWithCallback(this, this._swipeTimeout, Ci.nsITimer.TYPE_ONE_SHOT);
 				if (!this._directionChain) {
-					this.sourceNode = event.target;
-					this._lastX = event.screenX;
-					this._lastY = event.screenY;
+					this._startGesture(event);
 				}
 				var direction;
 				switch (event.direction) {
@@ -329,28 +345,6 @@ xdGestureHandler.prototype = {
 		return this._suppressAlt ? event.altKey : false;
 	},
 
-	_displayContextMenu: function FGH__displayContextMenu(event) {
-		if ("nsContextMenu" in this._drawArea.ownerDocument.defaultView) {
-			with (this._drawArea.ownerDocument.defaultView) {
-				if (!nsContextMenu.prototype._setTargetInternal) {
-					nsContextMenu.prototype._setTargetInternal = nsContextMenu.prototype.setTarget;
-					nsContextMenu.prototype.setTarget = function(aNode, aRangeParent, aRangeOffset) {
-						this._setTargetInternal(aNode, aRangeParent, this._rangeOffset);
-					};
-				}
-				nsContextMenu.prototype._rangeOffset = event.rangeOffset;
-			}
-		}
-		this._enableContextMenu(true);
-		var evt = event.originalTarget.ownerDocument.createEvent("MouseEvents");
-		evt.initMouseEvent(
-			"contextmenu", true, true, event.originalTarget.ownerDocument.defaultView, 0,
-			event.screenX, event.screenY, event.clientX, event.clientY,
-			false, false, false, false, 2, null
-		);
-		event.originalTarget.dispatchEvent(evt);
-	},
-
 	_enableContextMenu: function FGH__enableContextMenu(aEnable) {
 		var elt = this._drawArea.ownerDocument.getElementById("contentAreaContextMenu");
 		if (!elt)
@@ -373,17 +367,28 @@ xdGestureHandler.prototype = {
 			return;
 		event.preventDefault();
 		event.stopPropagation();
-		tabbar.advanceSelectedTab(event.detail < 0 ? -1 : 1, true);
+		tabbar.advanceSelectedTab(event.deltaY < 0 ? -1 : 1, true);
 	},
 
 	_startGesture: function FGH__startGesture(event) {
+		if (this._drawArea.localName == "tabbrowser")
+			this._isRemote = this._drawArea.mCurrentBrowser.getAttribute("remote") == "true";
 		this.sourceNode = event.target;
 		this._lastX = event.screenX;
 		this._lastY = event.screenY;
 		this._directionChain = "";
 		this._shouldFireContext = false;
-		if (this._trailEnabled)
-			this.createTrail(event);
+		if (!this._swipeTimer && this._trailEnabled)
+			this._createTrail();
+		if (this._isRemote) {
+			var zoom = this._drawArea.mCurrentBrowser.fullZoom;
+			this._gestureObserver.sendAsyncMessage("FireGestures:GestureStart", {
+				type: event.type, 
+				button: event.button, 
+				x: (event.screenX - this._drawArea.mCurrentBrowser.boxObject.screenX) / zoom, 
+				y: (event.screenY - this._drawArea.mCurrentBrowser.boxObject.screenY) / zoom, 
+			});
+		}
 	},
 
 	_progressGesture: function FGH__progressGesture(event) {
@@ -399,7 +404,7 @@ xdGestureHandler.prototype = {
 		else
 			direction = y < this._lastY ? "U" : "D";
 		if (this._trailEnabled)
-			this.drawTrail(this._lastX, this._lastY, x, y);
+			this._drawTrail(this._lastX, this._lastY, x, y);
 		this._lastX = x;
 		this._lastY = y;
 		if (this._state == STATE_KEYPRESS)
@@ -419,7 +424,7 @@ xdGestureHandler.prototype = {
 			this._lastExtraY = event.screenY;
 		}
 		if (this._state != STATE_KEYPRESS && this._trailEnabled)
-			this.eraseTrail();
+			this._eraseTrail();
 		if (!this.sourceNode)
 			this.sourceNode = event.target;
 		this._gestureObserver.onExtraGesture(event, aGestureType);
@@ -438,15 +443,16 @@ xdGestureHandler.prototype = {
 		this._isMouseDownM = false;
 		this._isMouseDownR = false;
 		this._clearTimeout();
-		if (this._trailEnabled)
-			this.eraseTrail();
+		if (!this._swipeTimer && this._trailEnabled)
+			this._eraseTrail();
 		if (this._directionChain) {
-			this._gestureObserver.onMouseGesture(event, this._directionChain);
+			var directionChain = this._directionChain;
+			this._directionChain = "";
+			this._gestureObserver.onMouseGesture(event, directionChain);
 			this._suppressContext = true;
 			this._shouldFireContext = false;
 		}
 		this.sourceNode = null;
-		this._directionChain = "";
 	},
 
 
@@ -481,9 +487,7 @@ xdGestureHandler.prototype = {
 				this._gestureObserver.onExtraGesture(null, "gesture-timeout");
 				break;
 			case this._swipeTimer: 
-				this._gestureObserver.onMouseGesture(null, this._directionChain);
-				this.sourceNode = null;
-				this._directionChain = "";
+				this._stopGesture();
 				this._swipeTimer = null;
 				break;
 		}
@@ -501,98 +505,81 @@ xdGestureHandler.prototype = {
 		this._stopGesture();
 	},
 
+	cancelMouseGesture: function FGH_cancelMouseGesture() {
+		this._directionChain = "";
+		this._stopGesture();
+	},
+
 
 
-	_trailDot: null,
 	_trailArea: null,
-	_trailLastDot: null,
+	_trailContext: null,
 	_trailOffsetX: 0,
 	_trailOffsetY: 0,
-	_trailZoom: 1,
 
-	createTrail: function FGH_createTrail(event) {
-		var win = event.view;
-		if (win.top.document instanceof Ci.nsIDOMHTMLDocument)
-			win = win.top;
-		else if (win.document instanceof Ci.nsIDOMHTMLDocument === false)
+	_createTrail: function FGH__createTrail() {
+		var doc = this._drawArea.ownerDocument;
+		var box = doc.documentElement.boxObject;
+		if (this._trailArea) {
+			this._trailArea.style.display = "-moz-box";
+			this._trailOffsetX = box.screenX;
+			this._trailOffsetY = box.screenY;
+			var canvas = this._trailArea.firstChild;
+			canvas.setAttribute("width",  box.width);
+			canvas.setAttribute("height", box.height);
 			return;
-		var doc = win.document;
-		var insertionNode = doc.documentElement ? doc.documentElement : doc;
-		var win = doc.defaultView;
-		this._trailZoom = this._gestureObserver.fullZoom;
-		this._trailOffsetX = (win.mozInnerScreenX - win.scrollX) * this._trailZoom;
-		this._trailOffsetY = (win.mozInnerScreenY - win.scrollY) * this._trailZoom;
-		this._trailArea = doc.createElementNS(HTML_NS, "xdTrailArea");
-		insertionNode.appendChild(this._trailArea);
-		this._trailDot = doc.createElementNS(HTML_NS, "xdTrailDot");
-		this._trailDot.style.width = this._trailSize + "px";
-		this._trailDot.style.height = this._trailSize + "px";
-		this._trailDot.style.background = this._trailColor;
-		this._trailDot.style.border = "0px";
-		this._trailDot.style.position = "absolute";
-		this._trailDot.style.zIndex = 2147483647;
+		}
+		var css = "-moz-user-focus: none !important;"
+		        + "-moz-user-select: none !important;"
+		        + "display: -moz-box !important;"
+		        + "box-sizing: border-box !important;"
+		        + "pointer-events: none !important;"
+		        + "margin: 0 !important;"
+		        + "padding: 0 !important;"
+		        + "width: 100% !important;"
+		        + "height: 100% !important;"
+		        + "border: none !important;"
+		        + "box-shadow: none !important;"
+		        + "overflow: hidden !important;"
+		        + "background: none !important;"
+		        + "opacity: 0.6 !important;"
+		        + "position: fixed !important;"
+		        + "top:  " + box.y + "px !important;"
+		        + "left: " + box.x + "px !important;"
+		        + "z-index: 2147483647 !important;";
+		this._trailArea = doc.createElement("hbox");
+		this._trailArea.id = "FireGesturesTrail";
+		this._trailArea.style.cssText = css;
+		this._trailOffsetX = box.screenX;
+		this._trailOffsetY = box.screenY;
+		var canvas = doc.createElementNS(HTML_NS, "canvas");
+		canvas.setAttribute("width",  box.width);
+		canvas.setAttribute("height", box.height);
+		this._trailArea.appendChild(canvas);
+		doc.documentElement.appendChild(this._trailArea);
+		this._trailContext = canvas.getContext("2d");
 	},
 
-	drawTrail: function FGH_drawTrail(x1, y1, x2, y2) {
+	_drawTrail: function FGH__drawTrail(x1, y1, x2, y2) {
 		if (!this._trailArea)
 			return;
-		var xMove = x2 - x1;
-		var yMove = y2 - y1;
-		var xDecrement = xMove < 0 ? 1 : -1;
-		var yDecrement = yMove < 0 ? 1 : -1;
-		x2 -= this._trailOffsetX;
-		y2 -= this._trailOffsetY;
-		if (Math.abs(xMove) >= Math.abs(yMove))
-			for (var i = xMove; i != 0; i += xDecrement)
-				this._strokeDot(x2 - i, y2 - Math.round(yMove * i / xMove));
-		else
-			for (var i = yMove; i != 0; i += yDecrement)
-				this._strokeDot(x2 - Math.round(xMove * i / yMove), y2 - i);
+		var context = this._trailContext;
+		context.strokeStyle = this._trailColor;
+		context.lineJoin = "round";
+		context.lineWidth = this._trailSize;
+		context.beginPath();
+		context.moveTo(x1 - this._trailOffsetX, y1 - this._trailOffsetY);
+		context.lineTo(x2 - this._trailOffsetX, y2 - this._trailOffsetY);
+		context.closePath();
+		context.stroke();
 	},
 
-	eraseTrail: function FGH_eraseTrail() {
-		if (this._trailArea && this._trailArea.parentNode) {
-			while (this._trailArea.lastChild)
-				this._trailArea.removeChild(this._trailArea.lastChild);
-			this._trailArea.parentNode.removeChild(this._trailArea);
-		}
-		this._trailDot = null;
-		this._trailArea = null;
-		this._trailLastDot = null;
-	},
-
-	_strokeDot: function FGH__strokeDot(x, y) {
-		if (this._trailArea.y == y && this._trailArea.h == this._trailSize) {
-			var newX = Math.min(this._trailArea.x, x);
-			var newW = Math.max(this._trailArea.x + this._trailArea.w, x + this._trailSize) - newX;
-			this._trailArea.x = newX;
-			this._trailArea.w = newW;
-			this._trailLastDot.style.left  = newX.toString() + "px";
-			this._trailLastDot.style.width = newW.toString() + "px";
+	_eraseTrail: function FGH__eraseTrail() {
+		if (!this._trailArea)
 			return;
-		}
-		else if (this._trailArea.x == x && this._trailArea.w == this._trailSize) {
-			var newY = Math.min(this._trailArea.y, y);
-			var newH = Math.max(this._trailArea.y + this._trailArea.h, y + this._trailSize) - newY;
-			this._trailArea.y = newY;
-			this._trailArea.h = newH;
-			this._trailLastDot.style.top    = newY.toString() + "px";
-			this._trailLastDot.style.height = newH.toString() + "px";
-			return;
-		}
-		if (this._trailZoom != 1) {
-			x = Math.floor(x / this._trailZoom);
-			y = Math.floor(y / this._trailZoom);
-		}
-		var dot = this._trailDot.cloneNode(true);
-		dot.style.left = x + "px";
-		dot.style.top = y + "px";
-		this._trailArea.x = x;
-		this._trailArea.y = y;
-		this._trailArea.w = this._trailSize;
-		this._trailArea.h = this._trailSize;
-		this._trailArea.appendChild(dot);
-		this._trailLastDot = dot;
+		var canvas = this._trailArea.firstChild;
+		this._trailContext.clearRect(0, 0, canvas.getAttribute("width"), canvas.getAttribute("height"));
+		this._trailArea.style.display = "none";
 	},
 
 };
